@@ -3,11 +3,14 @@ Test Generic Map
 """
 import re
 import tempfile
+import contextlib
 from unittest import mock
 
 import matplotlib.pyplot as plt
 import numpy as np
 import pytest
+from hypothesis import given, settings
+from matplotlib.figure import Figure
 from packaging import version
 
 import astropy.units as u
@@ -16,7 +19,6 @@ from astropy.coordinates import Latitude, SkyCoord
 from astropy.io import fits
 from astropy.io.fits.verify import VerifyWarning
 from astropy.tests.helper import assert_quantity_allclose
-from astropy.time import Time
 from astropy.visualization import wcsaxes
 
 import sunpy
@@ -25,11 +27,15 @@ import sunpy.data.test
 import sunpy.map
 import sunpy.sun
 from sunpy.coordinates import HeliographicCarrington, HeliographicStonyhurst, sun
+from sunpy.map.mapbase import GenericMap
 from sunpy.map.sources import AIAMap
+from sunpy.tests.helpers import figure_test
 from sunpy.time import parse_time
-from sunpy.util import SunpyDeprecationWarning, SunpyUserWarning
-from sunpy.util.exceptions import SunpyMetadataWarning
+from sunpy.util import SunpyUserWarning
+from sunpy.util.exceptions import SunpyDeprecationWarning, SunpyMetadataWarning
 from sunpy.util.metadata import ModifiedItem
+from .conftest import make_simple_map
+from .strategies import matrix_meta
 
 testpath = sunpy.data.test.rootdir
 
@@ -56,6 +62,8 @@ def test_wcs(aia171_test_map):
     wcs = aia171_test_map.wcs
     assert isinstance(wcs, astropy.wcs.WCS)
 
+    assert wcs.array_shape == aia171_test_map.data.shape
+
     assert all(wcs.wcs.crpix - 1 ==
                [aia171_test_map.reference_pixel.x.value, aia171_test_map.reference_pixel.y.value])
     assert u.allclose(wcs.wcs.cdelt * (u.Unit(wcs.wcs.cunit[0])/u.pix),
@@ -83,6 +91,19 @@ def test_wcs_cache(aia171_test_map):
     assert new_wcs.wcs.crpix[0] == new_crpix
 
 
+def test_obs_coord_cache(aia171_test_map):
+    coord1 = aia171_test_map.observer_coordinate
+    coord2 = aia171_test_map.observer_coordinate
+    assert coord1 is coord2
+
+    # Change metadata, and check that the coordinate changes
+    aia171_test_map.meta['haex_obs'] += 10
+    new_coord = aia171_test_map.observer_coordinate
+    assert new_coord.lon != coord2.lon
+    assert new_coord.lat != coord2.lat
+    assert new_coord.radius != coord2.radius
+
+
 def test_header_immutability(aia171_test_map):
     # Check that accessing the wcs of a map doesn't modify the meta data
     assert 'KEYCOMMENTS' in aia171_test_map.meta
@@ -94,11 +115,6 @@ def test_dtype(generic_map):
     assert generic_map.dtype == np.float64
 
 
-def test_size(generic_map):
-    with pytest.warns(SunpyDeprecationWarning, match='Use map.data.size instead'):
-        assert generic_map.size == 36 * u.pix
-
-
 def test_min(generic_map):
     assert generic_map.min() == 1
 
@@ -140,8 +156,37 @@ def test_nickname_set(generic_map):
     assert generic_map.nickname == 'hi'
 
 
-def test_date(generic_map):
-    assert isinstance(generic_map.date, Time)
+date_dict = {'DATE-AVG': parse_time('2020-01-01'),
+             'DATE-OBS': parse_time('2020-02-01'),
+             'DATE-BEG': parse_time('2020-03-01'),
+             'DATE-END': parse_time('2020-03-03')}
+date_begend = date_dict['DATE-BEG'] + (date_dict['DATE-END'] - date_dict['DATE-BEG']) / 2
+
+
+@pytest.mark.parametrize('keys, expected_date',
+                         ([['DATE-AVG', 'DATE-OBS', 'DATE-BEG', 'DATE-END'], date_dict['DATE-OBS']],
+                          [['DATE-AVG', 'DATE-BEG', 'DATE-END'], date_dict['DATE-AVG']],
+                          [['DATE-BEG', 'DATE-END'], date_begend],
+                          [['DATE-BEG'], date_dict['DATE-BEG']],
+                          [['DATE-END'], date_dict['DATE-END']],
+                          [[], 'now']
+                          ))
+def test_date(generic_map, keys, expected_date):
+    # Remove pre-existing date keys
+    for key in date_dict:
+        generic_map.meta.pop(key, None)
+    # Add new date keys
+    for key in keys:
+        generic_map.meta[key] = date_dict[key].isot
+    # Check date is the correct value
+    if expected_date == 'now':
+        expected_date = parse_time('now')
+        # Check equal to within a tolerance as parse_time('now') is run
+        # at slightly different times in .date and the line above
+        with pytest.warns(SunpyMetadataWarning, match='Missing metadata for observation time'):
+            assert generic_map.date - expected_date < 1*u.s
+    else:
+        assert generic_map.date == expected_date
 
 
 def test_date_scale(generic_map):
@@ -175,14 +220,12 @@ def test_rsun_meters(generic_map):
 
 
 def test_rsun_obs_without_rsun_ref(generic_map):
-    with pytest.warns(SunpyUserWarning,
-                      match='assuming the standard radius of the photosphere '
-                            'as seen from the observer distance'):
-        assert_quantity_allclose(generic_map.rsun_obs, sun.angular_radius(generic_map.date))
+    assert_quantity_allclose(generic_map.rsun_obs,
+                             sun.angular_radius(generic_map.date))
 
 
 def test_rsun_obs_with_rsun_ref(generic_map):
-    generic_map.meta['rsun_ref'] = sunpy.sun.constants.radius
+    generic_map.meta['rsun_ref'] = sunpy.sun.constants.radius.to_value(u.m)
     # The following should not raise a warning because we can calculate it exactly
     assert_quantity_allclose(generic_map.rsun_obs, sun.angular_radius(generic_map.date))
 
@@ -193,12 +236,12 @@ def test_coordinate_system(generic_map):
 
 def test_default_coordinate_system(generic_map):
     generic_map.meta.pop('ctype1')
-    with pytest.warns(SunpyUserWarning, match='Missing CTYPE1 from metadata'):
+    with pytest.warns(SunpyMetadataWarning, match='Missing CTYPE1 from metadata'):
         assert generic_map.coordinate_system == ('HPLN-TAN', 'HPLT-TAN')
 
     generic_map.meta.pop('ctype2')
     generic_map.meta['ctype1'] = 'HPLN-TAN'
-    with pytest.warns(SunpyUserWarning, match='Missing CTYPE2 from metadata'):
+    with pytest.warns(SunpyMetadataWarning, match='Missing CTYPE2 from metadata'):
         assert generic_map.coordinate_system == ('HPLN-TAN', 'HPLT-TAN')
 
 
@@ -325,6 +368,38 @@ def test_swap_cd():
     np.testing.assert_allclose(amap.rotation_matrix, np.array([[1., 0], [0, 1.]]))
 
 
+@pytest.mark.filterwarnings('ignore:Missing metadata for observer')
+def test_crota_scale():
+    # Test non-zero crota and unequal CDELT{1,2}
+    n = 6
+    data = np.ones([n, n], dtype=np.float64)
+    header = {
+        'CRVAL1': 0,
+        'CRVAL2': 0,
+        'CRPIX1': (n + 1) / 2,
+        'CRPIX2': (n + 1) / 2,
+        'NAXIS1': n,
+        'NAXIS2': n,
+        'CUNIT1': 'arcsec',
+        'CUNIT2': 'arcsec',
+        'CTYPE1': 'HPLN-TAN',
+        'CTYPE2': 'HPLT-TAN',
+        'DATE-OBS': '2020-01-01 00:00:00'
+    }
+
+    header.update({'CROTA2': 0, 'CDELT1': 1, 'CDELT2': 2})
+    map1 = sunpy.map.Map(data, header)
+    header.update({'CROTA2': 90, 'CDELT1': 2, 'CDELT2': 1})
+    map2 = sunpy.map.Map(data, header)
+
+    # Lower left coord
+    coord1 = map1.pixel_to_world(*(-0.5, -0.5) * u.pix)
+    # After rotating by 90 deg about the center of the map (CRPIX),
+    # this should map to the lower right coordinate
+    coord2 = map2.pixel_to_world(*(-0.5, n - 0.5) * u.pix)
+    assert coord1.separation(coord2) < 1e-6 * u.arcsec
+
+
 def test_world_to_pixel(generic_map):
     """Make sure conversion from data units to pixels is internally consistent"""
     test_pixel = generic_map.world_to_pixel(generic_map.reference_coordinate)
@@ -337,12 +412,10 @@ def test_world_to_pixel_error(generic_map):
         generic_map.world_to_pixel(1)
 
 
-@pytest.mark.parametrize('origin', [0, 1])
-def test_world_pixel_roundtrip(simple_map, origin):
+def test_world_pixel_roundtrip(simple_map):
     pix = 1 * u.pix, 1 * u.pix
-    with pytest.warns(SunpyDeprecationWarning, match='The origin argument is deprecated'):
-        coord = simple_map.pixel_to_world(*pix, origin=origin)
-        pix_roundtrip = simple_map.world_to_pixel(coord, origin=origin)
+    coord = simple_map.pixel_to_world(*pix)
+    pix_roundtrip = simple_map.world_to_pixel(coord)
 
     assert u.allclose(pix_roundtrip.x, pix[0], atol=1e-10 * u.pix)
     assert u.allclose(pix_roundtrip.y, pix[1], atol=1e-10 * u.pix)
@@ -603,8 +676,6 @@ def test_resample(simple_map, shape):
     # Should be the mean of [0,1,2,3,4,5,6,7,8,9]
     if shape == [1, 1]:
         assert resampled.data == np.array([[4]])
-    assert resampled.scale.axis1 == 3 / shape[0] * simple_map.scale.axis1
-    assert resampled.scale.axis2 == 3 / shape[1] * simple_map.scale.axis2
 
     # Check that the corner coordinates of the input and output are the same
     resampled_lower_left = resampled.pixel_to_world(-0.5 * u.pix, -0.5 * u.pix)
@@ -619,55 +690,58 @@ def test_resample(simple_map, shape):
     assert u.allclose(resampled_upper_left.Ty, original_upper_left.Ty)
 
 
-resample_test_data = [('linear', (100, 200) * u.pixel), ('neighbor', (128, 256) * u.pixel),
-                      ('nearest', (512, 128) * u.pixel), ('spline', (200, 200) * u.pixel)]
+resample_test_data = [('linear', (100, 200) * u.pixel, contextlib.nullcontext()),
+                      ('neighbor', (128, 256) * u.pixel, pytest.warns(SunpyDeprecationWarning)),
+                      ('nearest', (512, 128) * u.pixel, contextlib.nullcontext()),
+                      ('spline', (200, 200) * u.pixel, contextlib.nullcontext())]
 
 
-@pytest.mark.parametrize('sample_method, new_dimensions', resample_test_data)
-def test_resample_dimensions(generic_map, sample_method, new_dimensions):
+@pytest.mark.parametrize('sample_method, new_dimensions, cm', resample_test_data)
+def test_resample_dimensions(generic_map, sample_method, new_dimensions, cm):
     """Check that resampled map has expected dimensions."""
-    resampled_map = generic_map.resample(new_dimensions, method=sample_method)
-    assert resampled_map.dimensions[0] == new_dimensions[0]
-    assert resampled_map.dimensions[1] == new_dimensions[1]
+    with cm:
+        resampled_map = generic_map.resample(new_dimensions, method=sample_method)
+        assert resampled_map.dimensions[0] == new_dimensions[0]
+        assert resampled_map.dimensions[1] == new_dimensions[1]
 
 
-@pytest.mark.parametrize('sample_method, new_dimensions', resample_test_data)
-def test_resample_metadata(generic_map, sample_method, new_dimensions):
+@pytest.mark.parametrize('sample_method, new_dimensions, cm', resample_test_data)
+def test_resample_metadata(generic_map, sample_method, new_dimensions, cm):
     """
     Check that the resampled map has correctly adjusted metadata.
     """
-    resampled_map = generic_map.resample(new_dimensions, method=sample_method)
-    assert float(resampled_map.meta['cdelt1']) / generic_map.meta['cdelt1'] \
-        == float(generic_map.data.shape[1]) / resampled_map.data.shape[1]
-    assert float(resampled_map.meta['cdelt2']) / generic_map.meta['cdelt2'] \
-        == float(generic_map.data.shape[0]) / resampled_map.data.shape[0]
-    # TODO: we should really test the numbers here, not just that the correct
-    # header values have been modified. However, I am lazy and we have figure
-    # tests.
-    assert resampled_map.meta['crpix1'] != generic_map.meta['crpix1']
-    assert resampled_map.meta['crpix2'] != generic_map.meta['crpix2']
-    assert u.allclose(resampled_map.meta['crval1'], generic_map.meta['crval1'])
-    assert u.allclose(resampled_map.meta['crval2'], generic_map.meta['crval2'])
-    assert resampled_map.meta['naxis1'] == new_dimensions[0].value
-    assert resampled_map.meta['naxis2'] == new_dimensions[1].value
-    for key in generic_map.meta:
-        if key not in ('cdelt1', 'cdelt2', 'crpix1', 'crpix2', 'crval1',
-                       'crval2', 'naxis1', 'naxis2'):
-            assert resampled_map.meta[key] == generic_map.meta[key]
+    with cm:
+        resampled_map = generic_map.resample(new_dimensions, method=sample_method)
+        assert resampled_map.meta['cdelt1'] == generic_map.meta['cdelt1']
+        assert resampled_map.meta['cdelt2'] == generic_map.meta['cdelt2']
+        # TODO: we should really test the numbers here, not just that the correct
+        # header values have been modified. However, I am lazy and we have figure
+        # tests.
+        assert resampled_map.meta['crpix1'] != generic_map.meta['crpix1']
+        assert resampled_map.meta['crpix2'] != generic_map.meta['crpix2']
+        assert u.allclose(resampled_map.meta['crval1'], generic_map.meta['crval1'])
+        assert u.allclose(resampled_map.meta['crval2'], generic_map.meta['crval2'])
+        assert resampled_map.meta['naxis1'] == new_dimensions[0].value
+        assert resampled_map.meta['naxis2'] == new_dimensions[1].value
+        for key in generic_map.meta:
+            if key not in ('crpix1', 'crpix2', 'crval1',
+                           'crval2', 'naxis1', 'naxis2') and not key.startswith('pc'):
+                assert resampled_map.meta[key] == generic_map.meta[key]
 
 
-@pytest.mark.parametrize('method', ['neighbor', 'nearest', 'linear', 'spline'])
-def test_resample_simple_map(simple_map, method):
+@pytest.mark.parametrize('sample_method, new_dimensions, cm', resample_test_data)
+def test_resample_simple_map(simple_map, sample_method, new_dimensions, cm):
     # Put the reference pixel at the top-right of the bottom-left pixel
     simple_map.meta['crpix1'] = 1.5
     simple_map.meta['crpix2'] = 1.5
     assert list(simple_map.reference_pixel) == [0.5 * u.pix, 0.5 * u.pix]
     # Make the superpixel map
     new_dims = (9, 6) * u.pix
-    resamp_map = simple_map.resample(new_dims, method=method)
-    # Reference pixel should change, but reference coordinate should not
-    assert list(resamp_map.reference_pixel) == [2.5 * u.pix, 1.5 * u.pix]
-    assert resamp_map.reference_coordinate == simple_map.reference_coordinate
+    with cm:
+        resamp_map = simple_map.resample(new_dims, method=sample_method)
+        # Reference pixel should change, but reference coordinate should not
+        assert list(resamp_map.reference_pixel) == [2.5 * u.pix, 1.5 * u.pix]
+        assert resamp_map.reference_coordinate == simple_map.reference_coordinate
 
 
 def test_superpixel_simple_map(simple_map):
@@ -754,6 +828,38 @@ def test_superpixel_fractional_inputs(generic_map):
     assert super1.meta == super2.meta
 
 
+@pytest.mark.parametrize('method', ['resample', 'superpixel'])
+@settings(max_examples=10, deadline=1000)
+@given(pc=matrix_meta('pc'))
+def test_resample_rotated_map_pc(pc, method):
+    smap = make_simple_map()
+
+    smap.meta.update(pc)
+    # Check superpixel with a rotated map with unequal resampling
+    new_dims = (1, 2) * u.pix
+    new_map = getattr(smap, method)(new_dims)
+    # Coordinate of the lower left corner should not change
+    ll_pix = [-0.5, -0.5]*u.pix
+    assert smap.pixel_to_world(*ll_pix).separation(
+        new_map.pixel_to_world(*ll_pix)).to(u.arcsec) < 1e-8 * u.arcsec
+
+
+@pytest.mark.parametrize('method', ['resample', 'superpixel'])
+@settings(max_examples=10, deadline=1000)
+@given(cd=matrix_meta('cd'))
+def test_resample_rotated_map_cd(cd, method):
+    smap = make_simple_map()
+
+    smap.meta.update(cd)
+    # Check superpixel with a rotated map with unequal resampling
+    new_dims = (1, 2) * u.pix
+    new_map = getattr(smap, method)(new_dims)
+    # Coordinate of the lower left corner should not change
+    ll_pix = [-0.5, -0.5]*u.pix
+    assert smap.pixel_to_world(*ll_pix).separation(
+        new_map.pixel_to_world(*ll_pix)).to(u.arcsec) < 1e-8 * u.arcsec
+
+
 def test_superpixel_err(generic_map):
     with pytest.raises(ValueError, match="Offset is strictly non-negative."):
         generic_map.superpixel((2, 2) * u.pix, offset=(-2, 2) * u.pix)
@@ -766,6 +872,10 @@ def calc_new_matrix(angle):
 
 
 def test_rotate(aia171_test_map):
+    # The test map has big-endian floats, so we switch it to floats with native byte ordering
+    # Otherwise, errors can be raised by code that has been compiled
+    aia171_test_map._data = aia171_test_map.data.astype('float')
+
     rotated_map_1 = aia171_test_map.rotate(20 * u.deg)
     rotated_map_2 = rotated_map_1.rotate(20 * u.deg)
     np.testing.assert_allclose(rotated_map_1.rotation_matrix,
@@ -781,17 +891,14 @@ def test_rotate(aia171_test_map):
     np.testing.assert_allclose(rotated_map_2.data[0, 0], 0., atol=1e-7)
     assert rotated_map_2.mean() < rotated_map_1.mean() < aia171_test_map.mean()
 
-    rotated_map_3 = aia171_test_map.rotate(0 * u.deg, scale=1.5)
-    assert rotated_map_3.mean() > aia171_test_map.mean()
+    # A scaled-up map should have the same mean because the output map should be expanded
+    rotated_map_3 = aia171_test_map.rotate(0 * u.deg, order=3, scale=2)
+    np.testing.assert_allclose(aia171_test_map.mean(), rotated_map_3.mean(), rtol=1e-4)
 
-    # Mean and std should be equal when angle of rotation is integral multiple
-    # of 90 degrees for a square map
-    rotated_map_4 = aia171_test_map.rotate(90 * u.deg, scale=1.5)
-    np.testing.assert_allclose(rotated_map_3.mean(), rotated_map_4.mean(), rtol=1e-3)
-    np.testing.assert_allclose(rotated_map_3.std(), rotated_map_4.std(), rtol=1e-3)
-    rotated_map_5 = aia171_test_map.rotate(180 * u.deg, scale=1.5)
-    np.testing.assert_allclose(rotated_map_3.mean(), rotated_map_5.mean(), rtol=1e-3)
-    np.testing.assert_allclose(rotated_map_3.std(), rotated_map_5.std(), rtol=2e-3)
+    # Mean and std should be equal for a 90 degree rotation
+    rotated_map_4 = aia171_test_map.rotate(90 * u.deg, order=3, scale=2)
+    np.testing.assert_allclose(rotated_map_3.mean(), rotated_map_4.mean(), rtol=1e-10)
+    np.testing.assert_allclose(rotated_map_3.std(), rotated_map_4.std(), rtol=1e-10)
 
     # Rotation of a rectangular map by a large enough angle will change which dimension is larger
     aia171_test_map_crop = aia171_test_map.submap(
@@ -902,7 +1009,7 @@ def test_rotate_assumed_obstime():
     original = sunpy.map.Map(np.zeros((10, 10)), header)
 
     # Accessing the date makes the assumption of "now" for obstime
-    with pytest.warns(SunpyUserWarning, match="Missing metadata for observation time"):
+    with pytest.warns(SunpyMetadataWarning, match="Missing metadata for observation time"):
         original.date
 
     # The assumption has already been made, so no further warning should be emitted by rotate()
@@ -914,7 +1021,7 @@ def test_rotate_assumed_obstime():
     assert_quantity_allclose(rotated.reference_pixel.y, original.reference_pixel.y)
 
     # The returned map should also be missing observing time
-    with pytest.warns(SunpyUserWarning, match="Missing metadata for observation time"):
+    with pytest.warns(SunpyMetadataWarning, match="Missing metadata for observation time"):
         rotated.date
 
 
@@ -926,7 +1033,7 @@ def test_as_mpl_axes_aia171(aia171_test_map):
 
 def test_validate_meta(generic_map):
     """Check to see if_validate_meta displays an appropriate error"""
-    with pytest.warns(SunpyUserWarning) as w:
+    with pytest.warns(SunpyMetadataWarning) as w:
         bad_header = {
             'CRVAL1': 0,
             'CRVAL2': 0,
@@ -953,6 +1060,14 @@ def test_validate_meta(generic_map):
     assert 'waveunit'.upper() in str(w[0].message)
 
 
+def test_validate_non_spatial(generic_map):
+    generic_map.meta['cunit2'] = 'Angstrom'
+    err_msg = ("Map only supports spherical coordinate systems with angular units "
+               "(ie. equivalent to arcsec), but this map has units ['arcsec', 'Angstrom']")
+    with pytest.raises(sunpy.map.MapMetaValidationError, match=re.escape(err_msg)):
+        sunpy.map.Map(generic_map.data, generic_map.meta)
+
+
 # Heliographic Map Tests
 
 
@@ -978,38 +1093,6 @@ def test_hg_data_to_pix(heliographic_test_map):
     assert_quantity_allclose(out[0], 180 * u.pix)
     assert_quantity_allclose(out[1], 90 * u.pix)
 
-# Heliocentric Map Tests
-
-
-def test_hc_warn():
-    data = np.ones([6, 6], dtype=np.float64)
-    header = {
-        'CRVAL1': 0,
-        'CRVAL2': 0,
-        'CRPIX1': 5,
-        'CRPIX2': 5,
-        'CDELT1': 10,
-        'CDELT2': 10,
-        'CUNIT1': 'km',
-        'CUNIT2': 'km',
-        'CTYPE1': 'SOLX    ',
-        'CTYPE2': 'SOLY    ',
-        'PC1_1': 0,
-        'PC1_2': -1,
-        'PC2_1': 1,
-        'PC2_2': 0,
-        'NAXIS1': 6,
-        'NAXIS2': 6,
-        'date-obs': '1970/01/01T00:00:00',
-        'obsrvtry': 'Foo',
-        'detector': 'bar',
-        'wavelnth': 10,
-        'waveunit': 'm'
-    }
-
-    with pytest.warns(UserWarning):
-        sunpy.map.Map((data, header))
-
 # Dimension testing
 
 
@@ -1017,7 +1100,7 @@ def test_more_than_two_dimensions():
     """Checks to see if an appropriate error is raised when a FITS with more than two dimensions is
     loaded.  We need to load a >2-dim dataset with a TELESCOP header"""
 
-    # Data crudely represnts 4 stokes, 4 wavelengths with Y,X of 3 and 5.
+    # Data crudely represents 4 stokes, 4 wavelengths with Y,X of 3 and 5.
     bad_data = np.random.rand(4, 4, 3, 5)
     hdr = fits.Header()
     hdr['TELESCOP'] = 'XXX'
@@ -1203,8 +1286,7 @@ def test_contour_units(simple_map):
 def test_contour_input(simple_map):
     simple_map.meta['bunit'] = 'm'
 
-    with pytest.warns(SunpyDeprecationWarning,
-         match='Passing contour levels that are not an astropy Quantity'):
+    with pytest.raises(TypeError, match='The levels argument has no unit attribute'):
         simple_map.draw_contours(1.5)
     with pytest.raises(TypeError, match='The levels argument has no unit attribute'):
         simple_map.contour(1.5)
@@ -1234,12 +1316,6 @@ def test_print_map(generic_map):
     assert out_str in out_repr
 
 
-def test_fix_bitpix(aia171_test_map):
-    original = aia171_test_map.meta.pop('bitpix')
-    aia171_test_map._fix_bitpix()
-    assert aia171_test_map.meta['bitpix'] == original
-
-
 def test_parse_submap_quantity_inputs(aia171_test_map):
     bottom_left = (0, 0)*u.arcsec
     top_right = (200, 200)*u.arcsec
@@ -1293,14 +1369,14 @@ def test_meta_modifications(aia171_test_map):
     aiamap.meta['cdelt1'] = 20
 
     assert aiamap.meta.original_meta != aiamap.meta
-    assert aiamap.meta.added_items == {'bunit': 'ct'}
+    assert aiamap.meta.added_items == {}
     assert aiamap.meta.removed_items == {}
     assert aiamap.meta.modified_items == {'cdelt1': ModifiedItem(old_cdelt1, 20)}
 
     # Check that rotate doesn't modify the original metadata
     aiamap_rot = aiamap.rotate(30 * u.deg)
     assert aiamap_rot.meta.original_meta == aiamap.meta.original_meta
-    assert set(aiamap_rot.meta.added_items.keys()) == set(['bunit', 'pc1_1', 'pc1_2', 'pc2_1', 'pc2_2'])
+    assert set(aiamap_rot.meta.added_items.keys()) == set(['pc1_1', 'pc1_2', 'pc2_1', 'pc2_2'])
     assert set(aiamap_rot.meta.removed_items.keys()) == set(['crota2'])
     assert set(aiamap_rot.meta.modified_items) == set(['cdelt1', 'crpix1', 'crpix2', 'crval1'])
 
@@ -1345,3 +1421,114 @@ def test_rsun_meters_no_warning_for_hgs(heliographic_test_map):
 
     assert_quantity_allclose(heliographic_test_map.rsun_meters,
                              heliographic_test_map.meta['rsun_ref'] << u.m)
+
+
+@figure_test
+def test_rotation_rect_pixelated_data(aia171_test_map):
+    aia_map = sunpy.map.Map(aia171_test_map)
+    rect_map = aia_map.superpixel([2, 1] * u.pix, func=np.mean)
+    rect_rot_map = rect_map.rotate(30 * u.deg)
+    rect_rot_map.peek()
+
+
+@figure_test
+def test_derotating_nonpurerotation_pcij(aia171_test_map):
+    # The following map has a a PCij matrix that is not a pure rotation
+    weird_map = aia171_test_map.rotate(30*u.deg).superpixel([2, 1]*u.pix)
+
+    # De-rotating the map by its PCij matrix should result in a normal-looking map
+    derotated_map = weird_map.rotate()
+
+    fig = Figure(figsize=(8, 4))
+
+    ax1 = fig.add_subplot(121, projection=weird_map)
+    weird_map.plot(axes=ax1, title='Map with a non-pure-rotation PCij matrix')
+
+    ax2 = fig.add_subplot(122, projection=derotated_map)
+    derotated_map.plot(axes=ax2, title='De-rotated map')
+
+    return fig
+
+
+# This function is used in the arithmetic tests below
+def check_arithmetic_value_and_units(map_new, data_expected):
+    assert u.allclose(map_new.quantity, data_expected)
+    assert map_new.unit.is_equivalent(data_expected.unit)
+
+
+@pytest.mark.parametrize('value', [
+    10 * u.ct,
+    10 * u.mct,
+    u.Quantity([10], u.ct),
+    u.Quantity(np.random.rand(128), u.ct),
+    u.Quantity(np.random.rand(128, 128), u.ct),
+    u.Quantity(np.random.rand(128, 128), u.mct),
+])
+def test_map_arithmetic_addition_subtraction(aia171_test_map, value):
+    new_map = aia171_test_map + value
+    check_arithmetic_value_and_units(new_map, aia171_test_map.quantity + value)
+    new_map = value + aia171_test_map
+    check_arithmetic_value_and_units(new_map, value + aia171_test_map.quantity)
+    new_map = aia171_test_map - value
+    check_arithmetic_value_and_units(new_map, aia171_test_map.quantity - value)
+    new_map = value - aia171_test_map
+    check_arithmetic_value_and_units(new_map, value - aia171_test_map.quantity)
+
+
+@pytest.mark.parametrize('value', [
+    10 * u.ct,
+    u.Quantity([10], u.ct),
+    u.Quantity(np.random.rand(128), u.ct),
+    u.Quantity(np.random.rand(128, 128), u.ct),
+    10.0,
+    np.random.rand(128),
+    np.random.rand(128, 128),
+])
+def test_map_arithmetic_multiplication_division(aia171_test_map, value):
+    new_map = aia171_test_map * value
+    check_arithmetic_value_and_units(new_map, aia171_test_map.quantity * value)
+    new_map = value * aia171_test_map
+    check_arithmetic_value_and_units(new_map, value * aia171_test_map.quantity)
+    new_map = aia171_test_map / value
+    check_arithmetic_value_and_units(new_map, aia171_test_map.quantity / value)
+    with pytest.warns(RuntimeWarning, match='divide by zero encountered in true_divide'):
+        new_map = value / aia171_test_map
+        check_arithmetic_value_and_units(new_map, value / aia171_test_map.quantity)
+
+
+def test_map_arithmetic_pow(aia171_test_map):
+    new_map = aia171_test_map ** 2
+    check_arithmetic_value_and_units(new_map, aia171_test_map.quantity ** 2)
+
+
+def test_map_arithmetic_neg(aia171_test_map):
+    new_map = -aia171_test_map
+    check_arithmetic_value_and_units(new_map, -aia171_test_map.quantity)
+
+
+@pytest.mark.parametrize('value,warn_context', [
+    ('map', pytest.warns(RuntimeWarning)),
+    ('foobar', contextlib.nullcontext()),
+    (None, contextlib.nullcontext()),
+    (['foo', 'bar'], contextlib.nullcontext()),
+])
+def test_map_arithmetic_operations_raise_exceptions(aia171_test_map, value, warn_context):
+    value = aia171_test_map if value == 'map' else value
+    with pytest.raises(TypeError):
+        _ = aia171_test_map + value
+    with pytest.raises(TypeError):
+        _ = aia171_test_map * value
+    with pytest.raises(TypeError):
+        # A runtime warning is thrown when dividing by zero in the case of
+        # the map test
+        with warn_context:
+            _ = value / aia171_test_map
+
+
+def test_parse_fits_units():
+    # Check that we parse a BUNIT of G correctly.
+    out_unit = GenericMap._parse_fits_unit("Gauss")
+    assert out_unit == u.G
+
+    out_unit = GenericMap._parse_fits_unit("G")
+    assert out_unit == u.G
